#include <nSystem.h>
#include <stdio.h>

#include "nSysimp.h"
#include "fifoqueues.h"

typedef struct
{
  nTask owner;
  Queue mqueue;
  FifoQueue wqueue;
} *nMonitor;

nMonitor nMakeMonitor()
{
  nMonitor mon= (nMonitor)nMalloc(sizeof(*mon));
  mon->owner= NULL; // Task duenha del monitor
  mon->mqueue= MakeQueue(); // cola de espera para usar el monitor
  mon->wqueue= MakeFifoQueue(); // cola de espera de tareas que hacen wait
  return mon;
}

void nEnter(nMonitor mon)
{
  START_CRITICAL();

  // Si el monitor ya tiene un duenho
  if (mon->owner!=NULL)
  {
    // no debo pedir el monitor si ya lo tengo...
    if (mon->owner==current_task)
      nFatalError("nEnter", "Trying to own the same monitor twice\n");

    // cambio el estado de mi thread a esperando un monitor
    current_task->status= WAIT_MON;

    // agrego la tarea actual en la cola de espera del monitor
    PutTask(mon->mqueue, current_task);

    // cambio de contexto
    ResumeNextReadyTask();
  }

  // si llego aca es pq nadie tenia control del monitor
  // o pq alquien me puso en READY
  // finalmente tomamos el control del monitor
  mon->owner= current_task;

  END_CRITICAL();
}

void nExit(nMonitor mon)
{
  START_CRITICAL();

  // error
  if (mon->owner!=current_task)
    nFatalError("nExit", "This thread does not own this monitor\n");

  // dejamos el monitor
  mon->owner= NULL;

  // dejamos en ready task la tarea actual
  PushTask(ready_queue, current_task);

  // saca primera tarea de la cola de espera del monitor 
  // la marca ready
  // la deja en la cola ready
  ReadyFirstTask(mon->mqueue);

  // cambio de contexto
  ResumeNextReadyTask();
  
  END_CRITICAL();
}

void nWait(nMonitor mon)
{
  START_CRITICAL();
  
  // error
  if (mon->owner!=current_task)
    nFatalError("nWait", "This thread does not own this monitor\n");

  // liberamos el monitor
  mon->owner= NULL;

  // marcamos la tarea actual como esperando una condicion
  current_task->status= WAIT_COND;

  // agregamos la tarea actual a la cola de espera de wait
  // esta cola solo se "despierta" con un notifyAll
  PutObj(mon->wqueue, current_task);
 
  // resumir alguna tarea de la cola de espera del 
  // monitor
  ReadyFirstTask(mon->mqueue);

  // cambio de contexto
  ResumeNextReadyTask();

  // cuando vuelva aca
  // vuelvo a tomar el monitor
  // el wait vuelve con el control del monitor!
  mon->owner= current_task;

  END_CRITICAL();
}

void nNotifyAll(nMonitor mon)
{
  START_CRITICAL();

  // error
  if (mon->owner!=current_task)
    nFatalError("nNotifyAll", "This thread does not own this monitor\n");

  // traspasamos a todos los que esperan por un wait
  // a la cola de los que pueden tomar el monitor
  // los dejamos con estado de esperando al monitor
  while (!EmptyFifoQueue(mon->wqueue))
  {
    nTask task= (nTask)GetObj(mon->wqueue);
    task->status= WAIT_MON;
    PushTask(mon->mqueue, task);
  }

  // cuando se libere el monitor, estas tareas pueden "pelear"
  // por acceder nuevamente al control del monitor

  END_CRITICAL();
}

static void ReadyFirstTask(Queue queue)
{
  nTask task= GetTask(queue);
  if (task!=NULL)
  {
    task->status= READY;
    PushTask(ready_queue, task);
  } 
}

void nDestroyMonitor(nMonitor mon)
{
  DestroyQueue(mon->mqueue);
  DestroyFifoQueue(mon->wqueue);
  nFree(mon);
}
